#!/usr/bin/env python3
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.


import sys
if sys.version_info.major < 3:
    print("virt-manager is python3 only. Run this as ./setup.py")
    sys.exit(1)

import glob
import fnmatch
import os
import unittest

import distutils
import distutils.command.build
import distutils.command.install
import distutils.command.install_data
import distutils.command.install_egg_info
import distutils.command.sdist
import distutils.dist
import distutils.log
import distutils.sysconfig


sysprefix = distutils.sysconfig.get_config_var("prefix")


def _import_buildconfig():
    # A bit of crazyness to import the buildconfig file without importing
    # the rest of virtinst, so the build process doesn't require all the
    # runtime deps to be installed
    import warnings

    # 'imp' is deprecated. We use it elsewhere though too. Deal with using
    # the modern replacement when we replace all usage
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        import imp
        buildconfig = imp.load_source('buildconfig', 'virtinst/buildconfig.py')
        if "libvirt" in sys.modules:
            raise RuntimeError("Found libvirt in sys.modules. setup.py should "
                    "not import virtinst.")
        return buildconfig.BuildConfig


BuildConfig = _import_buildconfig()


# pylint: disable=attribute-defined-outside-init

_desktop_files = [
    ("share/applications", ["data/virt-manager.desktop.in"]),
]
_appdata_files = [
    ("share/appdata", ["data/virt-manager.appdata.xml.in"]),
]


def _generate_potfiles_in():
    def find(dirname, ext):
        ret = []
        for root, ignore, filenames in os.walk(dirname):
            for filename in fnmatch.filter(filenames, ext):
                ret.append(os.path.join(root, filename))
        ret.sort(key=lambda s: s.lower())
        return ret

    scripts = ["virt-manager", "virt-install",
               "virt-clone", "virt-convert", "virt-xml"]

    potfiles = "\n".join(scripts) + "\n\n"
    potfiles += "\n".join(find("virtManager", "*.py")) + "\n\n"
    potfiles += "\n".join(find("virtconv", "*.py")) + "\n\n"
    potfiles += "\n".join(find("virtinst", "*.py")) + "\n\n"

    for ignore, filelist in _desktop_files + _appdata_files:
        potfiles += "\n".join(filelist) + "\n"
    potfiles += "\n"

    potfiles += "\n".join(["[type: gettext/glade]" + f for
                          f in find("ui", "*.ui")]) + "\n\n"

    return potfiles


class my_build_i18n(distutils.command.build.build):
    """
    Add our desktop files to the list, saves us having to track setup.cfg
    """
    user_options = [
        ('merge-po', 'm', 'merge po files against template'),
    ]

    def initialize_options(self):
        self.merge_po = False
    def finalize_options(self):
        pass

    def run(self):
        potfiles = _generate_potfiles_in()
        potpath = "po/POTFILES.in"

        try:
            print("Writing %s" % potpath)
            open(potpath, "w").write(potfiles)
            self._run()
        finally:
            print("Removing %s" % potpath)
            os.unlink(potpath)

    def _run(self):
        # Borrowed from python-distutils-extra
        po_dir = "po"

        # Update po(t) files and print a report
        # We have to change the working dir to the po dir for intltool
        cmd = ["intltool-update",
               (self.merge_po and "-r" or "-p"), "-g", "virt-manager"]

        wd = os.getcwd()
        os.chdir("po")
        self.spawn(cmd)
        os.chdir(wd)
        max_po_mtime = 0
        for po_file in glob.glob("%s/*.po" % po_dir):
            lang = os.path.basename(po_file[:-3])
            mo_dir = os.path.join("build", "mo", lang, "LC_MESSAGES")
            mo_file = os.path.join(mo_dir, "virt-manager.mo")
            if not os.path.exists(mo_dir):
                os.makedirs(mo_dir)

            cmd = ["msgfmt", po_file, "-o", mo_file]
            po_mtime = os.path.getmtime(po_file)
            mo_mtime = (os.path.exists(mo_file) and
                        os.path.getmtime(mo_file)) or 0
            if po_mtime > max_po_mtime:
                max_po_mtime = po_mtime
            if po_mtime > mo_mtime:
                self.spawn(cmd)

            targetpath = os.path.join("share/locale", lang, "LC_MESSAGES")
            self.distribution.data_files.append((targetpath, (mo_file,)))

        # merge .in with translation
        for (file_set, switch) in [(_desktop_files, "-d"),
                                   (_appdata_files, "-x")]:
            for (target, files) in file_set:
                build_target = os.path.join("build", target)
                if not os.path.exists(build_target):
                    os.makedirs(build_target)

                files_merged = []
                for f in files:
                    if f.endswith(".in"):
                        file_merged = os.path.basename(f[:-3])
                    else:
                        file_merged = os.path.basename(f)

                    file_merged = os.path.join(build_target, file_merged)
                    cmd = ["intltool-merge", switch, po_dir, f,
                           file_merged]
                    mtime_merged = (os.path.exists(file_merged) and
                                    os.path.getmtime(file_merged)) or 0
                    mtime_file = os.path.getmtime(f)
                    if (mtime_merged < max_po_mtime or
                        mtime_merged < mtime_file):
                        # Only build if output is older than input (.po,.in)
                        self.spawn(cmd)
                    files_merged.append(file_merged)
                self.distribution.data_files.append((target, files_merged))


class my_build(distutils.command.build.build):
    """
    Create simple shell wrappers for /usr/bin/ tools to point to /usr/share
    Compile .pod file
    """

    def _make_bin_wrappers(self):
        cmds = ["virt-manager", "virt-install", "virt-clone",
                "virt-convert", "virt-xml"]
        if not os.path.exists("build"):
            os.mkdir("build")

        for app in cmds:
            sharepath = os.path.join(BuildConfig.prefix,
                "share", "virt-manager", app)

            wrapper = "#!/bin/sh\n\n"
            wrapper += "exec \"%s\" \"$@\"" % (sharepath)

            newpath = os.path.abspath(os.path.join("build", app))
            print("Generating %s" % newpath)
            open(newpath, "w").write(wrapper)


    def _make_man_pages(self):
        for path in glob.glob("man/*.pod"):
            base = os.path.basename(path)
            appname = os.path.splitext(base)[0]
            newpath = os.path.join(os.path.dirname(path),
                                   appname + ".1")

            print("Generating %s" % newpath)
            ret = os.system('pod2man '
                            '--center "Virtual Machine Manager" '
                            '--release %s --name %s '
                            '< %s > %s' % (BuildConfig.version,
                                           appname.upper(),
                                           path, newpath))
            if ret != 0:
                raise RuntimeError("Generating '%s' failed." % newpath)

        if os.system("grep -IRq 'Hey!' man/") == 0:
            raise RuntimeError("man pages have errors in them! "
                               "(grep for 'Hey!')")

    def _build_icons(self):
        for size in glob.glob(os.path.join("data/icons", "*")):
            for category in glob.glob(os.path.join(size, "*")):
                icons = []
                for icon in glob.glob(os.path.join(category, "*")):
                    icons.append(icon)
                if not icons:
                    continue

                category = os.path.basename(category)
                dest = ("share/icons/hicolor/%s/%s" %
                        (os.path.basename(size), category))
                if category != "apps":
                    dest = dest.replace("share/", "share/virt-manager/")

                self.distribution.data_files.append((dest, icons))


    def _make_bash_completion_files(self):
        scripts = ["virt-install", "virt-clone", "virt-convert", "virt-xml"]
        srcfile = "data/bash-completion.sh.in"
        builddir = "build/bash-completion/"
        if not os.path.exists(builddir):
            os.makedirs(builddir)

        instpaths = []
        for script in scripts:
            genfile = os.path.join(builddir, script)
            print("Generating %s" % genfile)
            src = open(srcfile, "r")
            dst = open(genfile, "w")
            dst.write(src.read().replace("::SCRIPTNAME::", script))
            dst.close()
            instpaths.append(genfile)

        bashdir = "share/bash-completion/completions/"
        self.distribution.data_files.append((bashdir, instpaths))


    def run(self):
        self._make_bin_wrappers()
        self._make_man_pages()
        self._build_icons()
        self._make_bash_completion_files()

        self.run_command("build_i18n")
        distutils.command.build.build.run(self)


class my_egg_info(distutils.command.install_egg_info.install_egg_info):
    """
    Disable egg_info installation, seems pointless for a non-library
    """
    def run(self):
        pass


class my_install(distutils.command.install.install):
    """
    Error if we weren't 'configure'd with the correct install prefix
    """
    def finalize_options(self):
        if self.prefix is None:
            if BuildConfig.prefix != sysprefix:
                print("Using configured prefix=%s instead of sysprefix=%s" % (
                    BuildConfig.prefix, sysprefix))
                self.prefix = BuildConfig.prefix
            else:
                print("Using sysprefix=%s" % sysprefix)
                self.prefix = sysprefix

        elif self.prefix != BuildConfig.prefix:
            print("Install prefix=%s doesn't match configure prefix=%s\n"
                  "Pass matching --prefix to 'setup.py configure'" %
                  (self.prefix, BuildConfig.prefix))
            sys.exit(1)

        distutils.command.install.install.finalize_options(self)


class my_install_data(distutils.command.install_data.install_data):
    def run(self):
        distutils.command.install_data.install_data.run(self)

        if not self.distribution.no_update_icon_cache:
            distutils.log.info("running gtk-update-icon-cache")
            icon_path = os.path.join(self.install_dir, "share/icons/hicolor")
            self.spawn(["gtk-update-icon-cache", "-q", "-t", icon_path])

        if not self.distribution.no_compile_schemas:
            distutils.log.info("compiling gsettings schemas")
            gschema_install = os.path.join(self.install_dir,
                "share/glib-2.0/schemas")
            self.spawn(["glib-compile-schemas", gschema_install])


class my_sdist(distutils.command.sdist.sdist):
    description = "Update virt-manager.spec; build sdist-tarball."

    def run(self):
        f1 = open('virt-manager.spec.in', 'r')
        f2 = open('virt-manager.spec', 'w')
        for line in f1:
            f2.write(line.replace('@VERSION@', BuildConfig.version))
        f1.close()
        f2.close()

        distutils.command.sdist.sdist.run(self)


###################
# Custom commands #
###################

class my_rpm(distutils.core.Command):
    user_options = []
    description = "Build src and noarch rpms."

    def initialize_options(self):
        pass
    def finalize_options(self):
        pass

    def run(self):
        """
        Run sdist, then 'rpmbuild' the tar.gz
        """
        self.run_command('sdist')
        os.system('rpmbuild -ta --clean dist/virt-manager-%s.tar.gz' %
                  BuildConfig.version)


class configure(distutils.core.Command):
    user_options = [
        ("prefix=", None, "installation prefix"),
        ("default-graphics=", None,
         "Default graphics type (spice or vnc) (default=spice)"),
        ("default-hvs=", None,
         "Comma separated list of hypervisors shown in 'Open Connection' "
         "wizard. (default=all hvs)"),

    ]
    description = "Configure the build, similar to ./configure"

    def finalize_options(self):
        pass

    def initialize_options(self):
        self.prefix = sysprefix
        self.default_graphics = None
        self.default_hvs = None


    def run(self):
        template = ""
        template += "[config]\n"
        template += "prefix = %s\n" % self.prefix
        if self.default_graphics is not None:
            template += "default_graphics = %s\n" % self.default_graphics
        if self.default_hvs is not None:
            template += "default_hvs = %s\n" % self.default_hvs

        open(BuildConfig.cfgpath, "w").write(template)
        print("Generated %s" % BuildConfig.cfgpath)


class TestBaseCommand(distutils.core.Command):
    user_options = [
        ('debug', 'd', 'Show debug output'),
        ('testverbose', None, 'Show verbose output'),
        ('coverage', 'c', 'Show coverage report'),
        ('regenerate-output', None, 'Regenerate test output'),
        ("only=", None,
         "Run only testcases whose name contains the passed string"),
        ("testfile=", None, "Specific test file to run (e.g "
                            "validation, storage, ...)"),
    ]

    def initialize_options(self):
        self.debug = 0
        self.testverbose = 0
        self.regenerate_output = 0
        self.coverage = 0
        self.only = None
        self._testfiles = []
        self._clistate = {}
        self._dir = os.getcwd()
        self.testfile = None
        self._force_verbose = False
        self._external_coverage = False

    def finalize_options(self):
        if self.only:
            # Can do --only many-devices to match on the cli testcase
            # for "virt-install-many-devices", despite the actual test
            # function name not containing any '-'
            self.only = self.only.replace("-", "_")

    def _find_tests_in_dir(self, dirname, excludes):
        testfiles = []
        for t in sorted(glob.glob(os.path.join(self._dir, dirname, '*.py'))):
            base = os.path.basename(t)
            if base in excludes + ["__init__.py"]:
                continue

            if self.testfile:
                check = os.path.basename(self.testfile)
                if base != check and base != (check + ".py"):
                    continue

            testfiles.append('.'.join(
                dirname.split("/") + [os.path.splitext(base)[0]]))

        if not testfiles:
            raise RuntimeError("--testfile didn't catch anything")
        return testfiles

    def run(self):
        cov = None
        if self.coverage:
            import coverage
            omit = ["/usr/*", "/*/tests/*", "*progress.py"]
            cov = coverage.coverage(omit=omit)
            cov.erase()
            if not self._external_coverage:
                cov.start()

        import tests as testsmodule
        testsmodule.utils.clistate.regenerate_output = bool(
                self.regenerate_output)
        testsmodule.utils.clistate.use_coverage = bool(cov)
        testsmodule.utils.clistate.debug = bool(self.debug)
        for key, val in self._clistate.items():
            setattr(testsmodule.utils.clistate, key, val)
        testsmodule.setup_logging()
        testsmodule.setup_cli_imports()

        # This makes the test runner report results before exiting from ctrl-c
        unittest.installHandler()

        tests = unittest.TestLoader().loadTestsFromNames(self._testfiles)
        if self.only:
            newtests = []
            for suite1 in tests:
                for suite2 in suite1:
                    for testcase in suite2:
                        if self.only in str(testcase):
                            newtests.append(testcase)

            if not newtests:
                print("--only didn't find any tests")
                sys.exit(1)
            tests = unittest.TestSuite(newtests)
            print("Running only:")
            for test in newtests:
                print("%s" % test)
            print("")

        verbosity = 1
        if self.debug or self.testverbose or self._force_verbose:
            verbosity = 2
        t = unittest.TextTestRunner(verbosity=verbosity)

        try:
            result = t.run(tests)
        except KeyboardInterrupt:
            sys.exit(1)

        if cov:
            if self._external_coverage:
                cov.load()
            else:
                cov.stop()
                cov.save()

        err = int(bool(len(result.failures) > 0 or
                       len(result.errors) > 0))
        if cov and not err:
            cov.report(show_missing=False, skip_covered=True)
        sys.exit(err)



class TestCommand(TestBaseCommand):
    description = "Runs a quick unit test suite"

    def run(self):
        '''
        Finds all the tests modules in tests/, and runs them.
        '''
        excludes = ["dist.py", "test_urls.py", "test_inject.py"]
        testfiles = self._find_tests_in_dir("tests", excludes)

        # Put clitest at the end, since it takes the longest
        for f in testfiles[:]:
            if "clitest" in f:
                testfiles.remove(f)
                testfiles.append(f)

        # Always want to put checkprops at the end to get accurate results
        for f in testfiles[:]:
            if "checkprops" in f:
                testfiles.remove(f)
                if not self.testfile:
                    testfiles.append(f)

        self._testfiles = testfiles
        TestBaseCommand.run(self)


class TestUI(TestBaseCommand):
    description = "Run UI dogtails tests"

    def run(self):
        self._testfiles = self._find_tests_in_dir("tests/uitests", [])
        self._force_verbose = True
        self._external_coverage = True
        TestBaseCommand.run(self)


class TestURLFetch(TestBaseCommand):
    description = "Test fetching kernels and isos from various distro trees"
    user_options = TestBaseCommand.user_options + [
        ('skip-libosinfo', None,
            "Don't use libosinfo for media/tree detection, "
            "Use our internal detection logic."),
        ('force-libosinfo', None,
            "Only use libosinfo for media/tree detection. This will skip "
            "some cases that are known not to work, like debian/ubuntu "
            "tree detection."),
        ('iso-only', None, "Only run iso tests"),
        ('url-only', None, "Only run url tests"),
    ]

    def initialize_options(self):
        TestBaseCommand.initialize_options(self)
        self.skip_libosinfo = 0
        self.force_libosinfo = 0
        self.iso_only = 0
        self.url_only = 0

    def finalize_options(self):
        TestBaseCommand.finalize_options(self)

    def run(self):
        self._testfiles = ["tests.test_urls"]
        self._clistate = {
            "url_iso_only": bool(self.iso_only),
            "url_only": bool(self.url_only),
            "url_skip_libosinfo": bool(self.skip_libosinfo),
            "url_force_libosinfo": bool(self.force_libosinfo),
        }
        TestBaseCommand.run(self)


class TestInitrdInject(TestBaseCommand):
    description = "Test initrd inject with real kernels, fetched from URLs"

    def run(self):
        self._testfiles = ["tests.test_inject"]
        TestBaseCommand.run(self)


class TestDist(TestBaseCommand):
    description = "Tests to run before cutting a release"

    def run(self):
        self._testfiles = ["tests.dist"]
        TestBaseCommand.run(self)


class CheckPylint(distutils.core.Command):
    user_options = [
        ("jobs=", "j", "use multiple processes to speed up Pylint"),
    ]
    description = "Check code using pylint and pycodestyle"

    def initialize_options(self):
        self.jobs = None

    def finalize_options(self):
        if self.jobs:
            self.jobs = int(self.jobs)

    def run(self):
        import pylint.lint
        import pycodestyle

        files = ["setup.py", "virt-install", "virt-clone",
                 "virt-convert", "virt-xml", "virt-manager",
                 "virtinst", "virtconv", "virtManager",
                 "tests"]

        try:
            import codespell_lib
            # pylint: disable=protected-access
            print("running codespell")
            codespell_lib._codespell.main(
                '-I', 'tests/codespell_dict.txt',
                '--skip', '*.pyc,*.zip,*.vmdk,*.iso,*.xml', *files)
        except ImportError:
            print("codespell is not installed. skipping...")
        except Exception as e:
            print("Error running codespell: %s" % e)

        output_format = sys.stdout.isatty() and "colorized" or "text"

        print("running pycodestyle")
        style_guide = pycodestyle.StyleGuide(
            config_file='setup.cfg',
            format="pylint",
            paths=files
        )
        report = style_guide.check_files()
        if style_guide.options.count:
            sys.stderr.write(str(report.total_errors) + '\n')

        print("running pylint")
        pylint_opts = [
            "--rcfile", "pylintrc",
            "--output-format=%s" % output_format,
        ]
        if self.jobs:
            pylint_opts += ["--jobs=%d" % self.jobs]

        pylint.lint.Run(files + pylint_opts)


class VMMDistribution(distutils.dist.Distribution):
    global_options = distutils.dist.Distribution.global_options + [
        ("no-update-icon-cache", None, "Don't run gtk-update-icon-cache"),
        ("no-compile-schemas", None, "Don't compile gsettings schemas"),
    ]

    def __init__(self, *args, **kwargs):
        self.no_update_icon_cache = False
        self.no_compile_schemas = False
        distutils.dist.Distribution.__init__(self, *args, **kwargs)


distutils.core.setup(
    name="virt-manager",
    version=BuildConfig.version,
    author="Cole Robinson",
    author_email="virt-tools-list@redhat.com",
    url="http://virt-manager.org",
    license="GPLv2+",

    # These wrappers are generated in our custom build command
    scripts=([
        "build/virt-manager",
        "build/virt-clone",
        "build/virt-install",
        "build/virt-convert",
        "build/virt-xml"]),

    data_files=[
        ("share/virt-manager/", [
            "virt-manager",
            "virt-install",
            "virt-clone",
            "virt-convert",
            "virt-xml",
        ]),
        ("share/glib-2.0/schemas",
         ["data/org.virt-manager.virt-manager.gschema.xml"]),
        ("share/virt-manager/ui", glob.glob("ui/*.ui")),

        ("share/man/man1", [
            "man/virt-manager.1",
            "man/virt-install.1",
            "man/virt-clone.1",
            "man/virt-convert.1",
            "man/virt-xml.1"
        ]),

        ("share/virt-manager/virtManager", glob.glob("virtManager/*.py")),
        ("share/virt-manager/virtManager/details",
            glob.glob("virtManager/details/*.py")),
        ("share/virt-manager/virtManager/device",
            glob.glob("virtManager/device/*.py")),
        ("share/virt-manager/virtManager/lib",
            glob.glob("virtManager/lib/*.py")),
        ("share/virt-manager/virtManager/object",
            glob.glob("virtManager/object/*.py")),
        ("share/virt-manager/virtinst",
            glob.glob("virtinst/*.py") + glob.glob("virtinst/build.cfg")),
        ("share/virt-manager/virtinst/devices",
            glob.glob("virtinst/devices/*.py")),
        ("share/virt-manager/virtinst/domain",
            glob.glob("virtinst/domain/*.py")),
        ("share/virt-manager/virtinst/install",
            glob.glob("virtinst/install/*.py")),
        ("share/virt-manager/virtconv",
            glob.glob("virtconv/*.py")),
    ],

    cmdclass={
        'build': my_build,
        'build_i18n': my_build_i18n,

        'sdist': my_sdist,
        'install': my_install,
        'install_data': my_install_data,
        'install_egg_info': my_egg_info,

        'configure': configure,

        'pylint': CheckPylint,
        'rpm': my_rpm,
        'test': TestCommand,
        'test_ui': TestUI,
        'test_urls': TestURLFetch,
        'test_initrd_inject': TestInitrdInject,
        'test_dist': TestDist,
    },

    distclass=VMMDistribution,
)



#
# Copyright 2006-2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

import os
import re

from .logger import log


_ETC_VCONSOLE = "/etc/vconsole.conf"
_KEYBOARD_DIR = "/etc/sysconfig/keyboard"
_XORG_CONF = "/etc/X11/xorg.conf"
_CONSOLE_SETUP_CONF = "/etc/default/console-setup"
_KEYBOARD_DEFAULT = "/etc/default/keyboard"


def _find_xkblayout(f):
    """
    Reads a keyboard layout from a file that defines an XKBLAYOUT
    variable, e.g. /etc/default/{keyboard,console-setup}.
    The format of these files is such that they can be 'sourced'
    in a shell script.

    Used for both /etc/default/keyboard and /etc/default/console-setup.
    The former is used by Debian 6.0 (Squeeze) and later.  The latter is
    used by older versions of Debian, and Ubuntu.
    """
    kt = None
    keymap_re = re.compile(r'\s*XKBLAYOUT="(?P<kt>[a-z-]+)"')
    for line in f:
        m = keymap_re.match(line)
        if m:
            kt = m.group('kt')
            break
    return kt


def _xorg_keymap(f):
    """
    Look in /etc/X11/xorg.conf for the host machine's keymap, and attempt to
    map it to a keymap supported by qemu
    """
    kt = None
    keymap_re = re.compile(r'\s*Option\s+"XkbLayout"\s+"(?P<kt>[a-z-]+)"')
    for line in f:
        m = keymap_re.match(line)
        if m:
            kt = m.group('kt')
            break
    return kt


def _sysconfig_keyboard(f):
    kt = None
    while 1:
        s = f.readline()
        if s == "":
            break
        s = s.strip()
        if s.startswith("#"):
            continue
        if (re.search("KEYMAP", s) is not None or
            re.search("KEYTABLE", s) is not None or
           (re.search("KEYBOARD", s) is not None and
            re.search("KEYBOARDTYPE", s) is None)):
            if '"' in s:
                delim = '"'
            elif '=' in s:
                delim = '='
            else:
                continue
            kt = s.split(delim)[1].strip()
            break
    return kt


def _default_keymap():
    """
    Look in various config files for the host machine's keymap, and attempt
    to map it to a keymap supported by qemu
    """
    # Set keymap to same as hosts
    default = "en-us"
    keymap = None

    kt = None

    if "VIRTINST_TEST_SUITE" in os.environ:
        return default

    for path, cb in [
        (_ETC_VCONSOLE, _sysconfig_keyboard),
        (_KEYBOARD_DIR, _sysconfig_keyboard),
        (_XORG_CONF, _xorg_keymap),
        (_KEYBOARD_DEFAULT, _find_xkblayout),
        (_CONSOLE_SETUP_CONF, _find_xkblayout)]:
        if not os.path.exists(path):
            continue

        try:
            f = open(path, "r")
            kt = cb(f)
            f.close()
            if kt:
                log.debug("Found keymap=%s in %s", kt, path)
                break
            log.debug("Didn't find keymap in '%s'", path)
        except Exception as e:
            log.debug("Error parsing '%s': %s", path, str(e))

    if kt is None:
        log.debug("Did not parse any usable keymapping.")
        return default

    kt = kt.lower()
    keymap = sanitize_keymap(kt)
    if not keymap:
        log.debug("Didn't match keymap '%s' in keytable!", kt)
        return default
    return keymap


##################
# Public helpers #
##################

# Host keytable entry : keymap name in qemu/xen
# Only use lower case entries: all lookups are .lower()'d
keytable = {
    "ar": "ar",
    "da": "da", "dk": "da",
    "de": "de",
    "de-ch": "de-ch",
    "en-gb": "en-gb", "gb": "en-gb", "uk": "en-gb",
    "en-us": "en-us", "us": "en-us",
    "es": "es",
    "et": "et",
    "fi": "fi", "se_fi": "fi",
    "fo": "fo",
    "fr": "fr",
    "fr-be": "fr-be", "be": "fr-be",
    "fr-ca": "fr-ca", "ca": "fr-ca",
    "fr-ch": "fr-ch", "fr_ch": "fr-ch",
    "hr": "hr",
    "hu": "hu",
    "is": "is",
    "it": "it",
    "ja": "ja", "jp106": "ja", "jp": "ja",
    "lt": "lt",
    "lv": "lv",
    "mk": "mk",
    "nl": "nl",
    "nl-be": "nl-be",
    "no": "no",
    "pl": "pl",
    "pt": "pt",
    "pt-br": "pt-br", "br": "pt-br", "br-abnt2": "pt-br",
    "ru": "ru",
    "sl": "sl",
    "sv": "sv",
    "th": "th",
    "tr": "tr",
}


_cached_keymap = None


def default_keymap():
    global _cached_keymap
    if _cached_keymap is None:
        _cached_keymap = _default_keymap()
    return _cached_keymap


def sanitize_keymap(kt):
    """
    Make sure the passed keymap roughly matches something in keytable
    """
    if kt.lower() in keytable:
        return keytable[kt]

    # Try a more intelligent lookup: strip out all '-' and '_', sort
    # the keytable keys putting the longest first, then compare
    # by string prefix
    clean_kt = kt.replace("-", "").replace("_", "")
    sorted_keys = sorted(list(keytable.keys()), key=len)

    for key in sorted_keys:
        origkey = key
        key = key.replace("-", "").replace("_", "")

        if clean_kt.startswith(key):
            return keytable[origkey]

    return None



#
# Copyright 2006-2007, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#
# Backends for the various URL types we support (http, https, ftp, local)

import ftplib
import io
import os
import subprocess
import tempfile
import urllib

import requests

from ..logger import log


##############################
# Mocking for the test suite #
##############################

def _in_testsuite():
    return "VIRTINST_TEST_SUITE" in os.environ


def _make_mock_url(url, filesyntax):
    if url.endswith("treeinfo"):
        # If the url is requesting treeinfo, give a fake treeinfo from
        # our testsuite data
        fn = ("%s/../../tests/cli-test-xml/fakerhel6tree/.treeinfo" %
                os.path.abspath(os.path.dirname(__file__)))
        abspath = os.path.abspath(fn)
    else:
        # Otherwise just copy this file
        abspath = os.path.abspath(__file__)

    if filesyntax:
        return "file://" + abspath
    return abspath


class _MockRequestsResponse:
    def __init__(self, url):
        log.debug("mocking requests session for url=%s", url)
        fn = _make_mock_url(url, filesyntax=False)
        self._content = open(fn).read()
        self.headers = {'content-length': len(self._content)}

    def raise_for_status(self):
        pass
    def iter_content(self, *args, **kwargs):
        dummy = args
        dummy = kwargs
        return [self._content.encode("utf-8")]


class _MockRequestsSession:
    def close(self):
        pass
    def head(self, url, *args, **kwargs):
        dummy = args
        dummy = kwargs
        return _MockRequestsResponse(url)
    def get(self, url, *args, **kwargs):
        dummy = args
        dummy = kwargs
        return _MockRequestsResponse(url)


class _MockFTPSession:
    def connect(self, *args, **kwargs):
        pass
    def login(self, *args, **kwargs):
        pass
    def voidcmd(self, *args, **kwargs):
        pass
    def quit(self, *args, **kwargs):
        pass
    def size(self, url):
        path = _make_mock_url(url, filesyntax=False)
        return os.path.getsize(path)


###########################
# Fetcher implementations #
###########################

class _URLFetcher(object):
    """
    This is a generic base class for fetching/extracting files from
    a media source, such as CD ISO, or HTTP/HTTPS/FTP server
    """
    _block_size = 16384
    _is_iso = False

    def __init__(self, location, scratchdir, meter):
        self.location = location
        self.scratchdir = scratchdir
        self.meter = meter

        log.debug("Using scratchdir=%s", scratchdir)
        self._prepare()


    ####################
    # Internal helpers #
    ####################

    def _make_full_url(self, filename):
        """
        Generate a full fetchable URL from the passed filename, which
        is relative to the self.location
        """
        if self._is_iso:
            return os.path.join("/", filename)
        if not filename:
            return self.location
        return os.path.join(self.location, filename)

    def _grabURL(self, filename, fileobj, fullurl=None):
        """
        Download the filename from self.location, and write contents to
        fileobj
        """
        if fullurl:
            url = fullurl
        else:
            url = self._make_full_url(filename)

        try:
            urlobj, size = self._grabber(url)
        except Exception as e:
            raise ValueError(_("Couldn't acquire file %s: %s") %
                               (url, str(e)))

        log.debug("Fetching URI: %s", url)
        self.meter.start(
            text=_("Retrieving file %s...") % os.path.basename(filename),
            size=size)

        total = self._write(urlobj, fileobj)
        self.meter.end(total)

    def _write(self, urlobj, fileobj):
        """
        Write the contents of urlobj to python file like object fileobj
        """
        total = 0
        while 1:
            buff = urlobj.read(self._block_size)
            if not buff:
                break
            fileobj.write(buff)
            total += len(buff)
            self.meter.update(total)
        return total

    def _grabber(self, url):
        """
        Returns the urlobj, size for the passed URL. urlobj is whatever
        data needs to be passed to self._write
        """
        raise NotImplementedError("must be implemented in subclass")


    ##############
    # Public API #
    ##############

    def is_iso(self):
        """
        If this is a fetcher for local CDROM/ISO
        """
        return self._is_iso

    def _prepare(self):
        """
        Perform any necessary setup
        """

    def _cleanup(self):
        """
        Perform any necessary cleanup
        """

    def __del__(self):
        self._cleanup()

    def can_access(self):
        """
        Return True if the location URL seems to be valid
        """
        return True

    def _hasFile(self, url):
        raise NotImplementedError("Must be implemented in subclass")

    def hasFile(self, filename):
        """
        Return True if self.location has the passed filename
        """
        url = self._make_full_url(filename)
        ret = self._hasFile(url)
        log.debug("hasFile(%s) returning %s", url, ret)
        return ret

    def acquireFile(self, filename, fullurl=None):
        """
        Grab the passed filename from self.location and save it to
        a temporary file, returning the temp filename
        """
        # pylint: disable=redefined-variable-type
        fn = None
        try:
            fileobj = tempfile.NamedTemporaryFile(
                prefix="virtinst-", suffix="-" + os.path.basename(filename),
                dir=self.scratchdir, delete=False)
            fn = fileobj.name

            self._grabURL(filename, fileobj, fullurl=fullurl)
            log.debug("Saved file to %s", fn)
            return fn
        except:  # noqa
            if fn and os.path.exists(fn):  # pragma: no cover
                os.unlink(fn)  # pragma: no cover
            raise  # pragma: no cover

    def acquireFileContent(self, filename):
        """
        Grab the passed filename from self.location and return it as a string
        """
        fileobj = io.BytesIO()
        self._grabURL(filename, fileobj)
        return fileobj.getvalue().decode("utf-8")


class _HTTPURLFetcher(_URLFetcher):
    _session = None

    def _prepare(self):
        if _in_testsuite():
            self._session = _MockRequestsSession()
        else:
            self._session = requests.Session()

    def _cleanup(self):
        if self._session:
            try:
                self._session.close()
            except Exception:  # pragma: no cover
                log.debug("Error closing requests.session", exc_info=True)
        self._session = None

    def can_access(self):
        return self.hasFile("")

    def _hasFile(self, url):
        """
        We just do a HEAD request to see if the file exists
        """
        try:
            response = self._session.head(url, allow_redirects=True)
            response.raise_for_status()
        except Exception as e:  # pragma: no cover
            log.debug("HTTP hasFile request failed: %s", str(e))
            return False
        return True

    def _grabber(self, url):
        """
        Use requests for this
        """
        response = self._session.get(url, stream=True)
        response.raise_for_status()
        try:
            size = int(response.headers.get('content-length'))
        except Exception:  # pragma: no cover
            size = None
        return response, size

    def _write(self, urlobj, fileobj):
        """
        The requests object doesn't have a file-like read() option, so
        we need to implement it ourselves
        """
        total = 0
        for data in urlobj.iter_content(chunk_size=self._block_size):
            fileobj.write(data)
            total += len(data)
            self.meter.update(total)
        return total


class _FTPURLFetcher(_URLFetcher):
    _ftp = None

    def _prepare(self):
        if self._ftp:
            return  # pragma: no cover

        try:
            parsed = urllib.parse.urlparse(self.location)
            if _in_testsuite():
                self._ftp = _MockFTPSession()
            else:
                self._ftp = ftplib.FTP()
            username = urllib.parse.unquote(parsed.username or '')
            password = urllib.parse.unquote(parsed.password or '')
            self._ftp.connect(parsed.hostname, parsed.port or 0)
            self._ftp.login(username, password)
            # Force binary mode
            self._ftp.voidcmd("TYPE I")
        except Exception as e:  # pragma: no cover
            raise ValueError(_("Opening URL %s failed: %s.") %
                              (self.location, str(e)))

    def _grabber(self, url):
        """
        Use urllib and ftplib to grab the file
        """
        if _in_testsuite():
            url = _make_mock_url(url, filesyntax=True)
        request = urllib.request.Request(url)
        urlobj = urllib.request.urlopen(request)
        size = self._ftp.size(urllib.parse.urlparse(url)[2])
        return urlobj, size


    def _cleanup(self):
        if not self._ftp:
            return  # pragma: no cover

        try:
            self._ftp.quit()
        except Exception:  # pragma: no cover
            log.debug("Error quitting ftp connection", exc_info=True)

        self._ftp = None

    def _hasFile(self, url):
        path = urllib.parse.urlparse(url)[2]

        try:
            try:
                # If it's a file
                self._ftp.size(path)
            except ftplib.all_errors:
                # If it's a dir
                self._ftp.cwd(path)
        except ftplib.all_errors as e:  # pragma: no cover
            log.debug("FTP hasFile: couldn't access %s: %s",
                          url, str(e))
            return False

        return True


class _LocalURLFetcher(_URLFetcher):
    """
    For grabbing files from a local directory
    """
    def _hasFile(self, url):
        return os.path.exists(url)

    def _grabber(self, url):
        urlobj = open(url, "rb")
        size = os.path.getsize(url)
        return urlobj, size


class _ISOURLFetcher(_URLFetcher):
    _cache_file_list = None
    _is_iso = True

    def _grabber(self, url):
        """
        Use isoinfo to grab the file
        """
        if not self._hasFile(url):
            raise RuntimeError("isoinfo didn't find file=%s" % url)

        cmd = ["isoinfo", "-J", "-i", self.location, "-x", url]

        log.debug("Running isoinfo: %s", cmd)
        output = subprocess.check_output(cmd)

        return io.BytesIO(output), len(output)

    def _hasFile(self, url):
        """
        Use isoinfo to list and search for the file
        """
        if not self._cache_file_list:
            cmd = ["isoinfo", "-J", "-i", self.location, "-f"]

            log.debug("Running isoinfo: %s", cmd)
            output = subprocess.check_output(cmd, stderr=subprocess.DEVNULL)

            self._cache_file_list = output.splitlines(False)

        return url.encode("ascii") in self._cache_file_list


class DirectFetcher(_URLFetcher):
    def _make_full_url(self, filename):
        return filename

    def acquireFile(self, filename, fullurl=None):
        fullurl = filename
        filename = os.path.basename(filename)
        fetcher = fetcherForURI(fullurl, self.scratchdir, self.meter, direct=True)
        return fetcher.acquireFile(filename, fullurl)  # pylint: disable=protected-access

    def _hasFile(self, url):
        return True

    def _grabber(self, url):
        raise RuntimeError(  # pragma: no cover
                "DirectFetcher shouldn't be used for file access.")


def fetcherForURI(uri, scratchdir, meter, direct=False):
    if uri.startswith("http://") or uri.startswith("https://"):
        fclass = _HTTPURLFetcher
    elif uri.startswith("ftp://"):
        fclass = _FTPURLFetcher
    elif direct or os.path.isdir(uri):
        # Pointing to a local tree
        fclass = _LocalURLFetcher
    else:
        # Pointing to a path (e.g. iso), or a block device (e.g. /dev/cdrom)
        fclass = _ISOURLFetcher
    return fclass(uri, scratchdir, meter)



#
# Copyright 2017 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.


from .device import Device
from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


class _DeviceMemoryTarget(XMLBuilder):
    XML_NAME = "target"

    size = XMLProperty("./size", is_int=True)
    node = XMLProperty("./node", is_int=True)
    label_size = XMLProperty("./label/size", is_int=True)


class _DeviceMemorySource(XMLBuilder):
    XML_NAME = "source"

    pagesize = XMLProperty("./pagesize", is_int=True)
    nodemask = XMLProperty("./nodemask")
    path = XMLProperty("./path")


class DeviceMemory(Device):
    XML_NAME = "memory"

    MODEL_DIMM = "dimm"
    MODEL_NVDIMM = "nvdimm"
    models = [MODEL_DIMM, MODEL_NVDIMM]

    ACCESS_SHARED = "shared"
    ACCESS_PRIVATE = "private"
    accesses = [ACCESS_SHARED, ACCESS_PRIVATE]

    model = XMLProperty("./@model")
    access = XMLProperty("./@access")

    source = XMLChildProperty(_DeviceMemorySource, is_single=True)
    target = XMLChildProperty(_DeviceMemoryTarget, is_single=True)



#
# Common code for all guests
#
# Copyright 2006-2009, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from . import unattended
from .installertreemedia import InstallerTreeMedia
from .installerinject import perform_cdrom_injections
from ..domain import DomainOs
from ..devices import DeviceDisk
from ..osdict import OSDB
from ..logger import log
from .. import progress


def _make_testsuite_path(path):
    return os.path.join("/VIRTINST-TESTSUITE",
            os.path.basename(path).split("-", 2)[-1])


class Installer(object):
    """
    Class for kicking off VM installs. The VM is set up separately in a Guest
    instance. This class tracks the install media/bootdev choice, alters the
    Guest XML, boots it for the install, then saves the post install XML
    config. The Guest is passed in via start_install, only install media
    selection is done at __init__ time

    :param cdrom: Path to a cdrom device or iso. Maps to virt-install --cdrom
    :param location: An install tree URI, local directory, or ISO/CDROM path.
        Largely handled by installtreemedia helper class. Maps to virt-install
        --location
    :param location_kernel: URL pointing to a kernel to fetch, or a relative
        path to indicate where the kernel is stored in location
    :param location_initrd: location_kernel, but pointing to an initrd
    :param install_kernel: Kernel to install off of
    :param install_initrd: Initrd to install off of
    :param install_kernel_args: Kernel args <cmdline> to use. This overwrites
        whatever the installer might request, unlike extra_args which will
        append arguments.
    :param no_install: If True, this installer specifically does not have
        an install phase. We are just using it to create the initial XML.
    """
    def __init__(self, conn, cdrom=None, location=None, install_bootdev=None,
            location_kernel=None, location_initrd=None,
            install_kernel=None, install_initrd=None, install_kernel_args=None,
            no_install=None):
        self.conn = conn

        # Entry point for virt-manager 'Customize' wizard to change autostart
        self.autostart = False

        self._install_cdrom_device_added = False
        self._unattended_install_cdrom_device = None
        self._tmpfiles = []
        self._defaults_are_set = False
        self._unattended_data = None

        self._install_bootdev = install_bootdev
        self._no_install = no_install

        self._treemedia = None
        self._treemedia_bootconfig = None
        self._cdrom = None
        if cdrom:
            cdrom = InstallerTreeMedia.validate_path(self.conn, cdrom)
            self._cdrom = cdrom
            self._install_bootdev = "cdrom"
        elif (location or location_kernel or location_initrd or
              install_kernel or install_initrd):
            self._treemedia = InstallerTreeMedia(self.conn, location,
                    location_kernel, location_initrd,
                    install_kernel, install_initrd, install_kernel_args)


    ###################
    # Private helpers #
    ###################

    def _cdrom_path(self):
        if self._treemedia:
            return self._treemedia.cdrom_path()
        return self._cdrom

    def _add_install_cdrom_device(self, guest):
        if self._install_cdrom_device_added:
            return  # pragma: no cover
        if not bool(self._cdrom_path()):
            return

        dev = DeviceDisk(self.conn)
        dev.device = dev.DEVICE_CDROM
        dev.path = self._cdrom_path()
        dev.sync_path_props()
        dev.validate()
        self._install_cdrom_device_added = True

        # Insert the CDROM before any other CDROM, so boot=cdrom picks
        # it as the priority
        for idx, disk in enumerate(guest.devices.disk):
            if disk.is_cdrom():
                guest.devices.add_child(dev, idx=idx)
                return
        guest.add_device(dev)

    def _remove_install_cdrom_media(self, guest):
        if not self._install_cdrom_device_added:
            return
        if guest.osinfo.is_windows():
            # Keep media attached for windows which has a multi stage install
            return
        for disk in guest.devices.disk:
            if disk.is_cdrom() and disk.path == self._cdrom_path():
                disk.path = None
                disk.sync_path_props()
                break

    def _add_unattended_install_cdrom_device(self, guest, location):
        if self._unattended_install_cdrom_device:
            return  # pragma: no cover
        dev = DeviceDisk(self.conn)
        dev.device = dev.DEVICE_CDROM
        dev.path = location
        dev.sync_path_props()
        dev.validate()
        dev.set_defaults(guest)
        self._unattended_install_cdrom_device = dev
        guest.add_device(self._unattended_install_cdrom_device)

        if self.conn.in_testsuite():
            # pylint: disable=protected-access
            # Hack to set just the XML path differently for the test suite.
            # Setting this via regular 'path' will error that it doesn't exist
            dev._source_file = _make_testsuite_path(location)

    def _remove_unattended_install_cdrom_device(self, guest):
        dummy = guest
        if not self._unattended_install_cdrom_device:
            return

        self._unattended_install_cdrom_device.path = None
        self._unattended_install_cdrom_device.sync_path_props()

    def _build_boot_order(self, guest, bootdev):
        bootorder = [bootdev]

        # If guest has an attached disk, always have 'hd' in the boot
        # list, so disks are marked as bootable/installable (needed for
        # windows virtio installs, and booting local disk from PXE)
        for disk in guest.devices.disk:
            if disk.device == disk.DEVICE_DISK:
                bootdev = "hd"
                if bootdev not in bootorder:
                    bootorder.append(bootdev)
                break
        return bootorder

    def _can_set_guest_bootorder(self, guest):
        return (not guest.os.is_container() and
            not guest.os.kernel and
            not any([d.boot.order for d in guest.devices.get_all()]))

    def _alter_treemedia_bootconfig(self, guest):
        if not self._treemedia:
            return

        kernel, initrd, kernel_args = self._treemedia_bootconfig
        if kernel:
            guest.os.kernel = (self.conn.in_testsuite() and
                    _make_testsuite_path(kernel) or kernel)
        if initrd:
            guest.os.initrd = (self.conn.in_testsuite() and
                    _make_testsuite_path(initrd) or initrd)
        if kernel_args:
            guest.os.kernel_args = kernel_args

    def _alter_bootconfig(self, guest):
        """
        Generate the portion of the guest xml that determines boot devices
        and parameters. (typically the <os></os> block)

        :param guest: Guest instance we are installing
        """
        guest.on_reboot = "destroy"
        self._alter_treemedia_bootconfig(guest)

        bootdev = self._install_bootdev
        if bootdev and self._can_set_guest_bootorder(guest):
            guest.os.bootorder = self._build_boot_order(guest, bootdev)
        else:
            guest.os.bootorder = []

    def _alter_install_resources(self, guest, meter):
        """
        Sets the appropriate amount of ram needed when performing a "network"
        based installation

        :param guest: Guest instance we are installing
        """
        if not self._treemedia:
            return
        if not self._treemedia.requires_internet(guest, meter):
            return

        ram = guest.osinfo.get_network_install_required_ram(guest)
        ram = (ram or 0) // 1024
        if ram > guest.currentMemory:
            log.warning(_("Overriding memory to %s MiB needed for %s "
                "network install."), ram // 1024, guest.osinfo.name)
            guest.currentMemory = ram


    ##########################
    # Internal API overrides #
    ##########################

    def _prepare_unattended_data(self, guest, script):
        expected_filename = script.get_expected_filename()
        unattended_cmdline = script.generate_cmdline()
        log.debug("Generated unattended cmdline: %s", unattended_cmdline)

        scriptpath = script.write()
        self._tmpfiles.append(scriptpath)

        iso = perform_cdrom_injections([(scriptpath, expected_filename)],
                InstallerTreeMedia.make_scratchdir(guest))
        self._tmpfiles.append(iso)
        self._add_unattended_install_cdrom_device(guest, iso)

    def _prepare_unattended_script(self, guest, meter):
        url = None
        if self._treemedia:
            if self._treemedia.is_network_url():
                url = self.location
            os_media = self._treemedia.get_os_media(guest, meter)
            injection_method = "initrd"
        else:
            if self.conn.is_remote():
                raise RuntimeError("Unattended method=cdrom installs are "
                        "not yet supported for remote connections.")
            if not guest.osinfo.is_windows():
                log.warning("Attempting unattended method=cdrom injection "
                        "for a non-windows OS. If this doesn't work, try "
                        "passing install media to --location")
            osguess = OSDB.guess_os_by_iso(self.cdrom)
            os_media = osguess[1] if osguess else None
            injection_method = "cdrom"

        return unattended.prepare_install_script(
                guest, self._unattended_data, url, os_media, injection_method)

    def _prepare(self, guest, meter):
        unattended_script = None
        if self._unattended_data:
            unattended_script = self._prepare_unattended_script(guest, meter)

        if self._treemedia:
            self._treemedia_bootconfig = self._treemedia.prepare(guest, meter,
                    unattended_script)

        elif unattended_script:
            self._prepare_unattended_data(guest, unattended_script)

    def _cleanup(self, guest):
        if self._treemedia:
            self._treemedia.cleanup(guest)

        for f in self._tmpfiles:
            log.debug("Removing %s", str(f))
            os.unlink(f)

    def _get_postinstall_bootdev(self, guest):
        if self.cdrom and self._no_install:
            return DomainOs.BOOT_DEVICE_CDROM

        if self._install_bootdev:
            if any([d for d in guest.devices.disk
                    if d.device == d.DEVICE_DISK]):
                return DomainOs.BOOT_DEVICE_HARDDISK
            return self._install_bootdev

        device = guest.devices.disk and guest.devices.disk[0].device or None
        if device == DeviceDisk.DEVICE_DISK:
            return DomainOs.BOOT_DEVICE_HARDDISK
        elif device == DeviceDisk.DEVICE_CDROM:
            return DomainOs.BOOT_DEVICE_CDROM
        elif device == DeviceDisk.DEVICE_FLOPPY:
            return DomainOs.BOOT_DEVICE_FLOPPY
        return DomainOs.BOOT_DEVICE_HARDDISK


    ##############
    # Public API #
    ##############

    @property
    def location(self):
        if self._treemedia:
            return self._treemedia.location

    @property
    def cdrom(self):
        return self._cdrom

    def set_initrd_injections(self, initrd_injections):
        if not self._treemedia:
            raise RuntimeError("Install method does not support "
                    "initrd injections.")
        self._treemedia.set_initrd_injections(initrd_injections)

    def set_extra_args(self, extra_args):
        if not self._treemedia:
            raise RuntimeError("Kernel arguments are only supported with "
                    "location or kernel installs.")
        self._treemedia.set_extra_args(extra_args)

    def set_install_defaults(self, guest):
        """
        Allow API users to set defaults ahead of time if they want it.
        Used by vmmDomainVirtinst so the 'Customize before install' dialog
        shows accurate values.

        If the user doesn't explicitly call this, it will be called by
        start_install()
        """
        if self._defaults_are_set:
            return

        self._add_install_cdrom_device(guest)

        if not guest.os.bootorder and self._can_set_guest_bootorder(guest):
            bootdev = self._get_postinstall_bootdev(guest)
            guest.os.bootorder = self._build_boot_order(guest, bootdev)

        guest.set_defaults(None)
        self._defaults_are_set = True

    def get_search_paths(self, guest):
        """
        Return a list of paths that the hypervisor will need search access
        for to perform this install.
        """
        search_paths = []
        if self._treemedia:
            search_paths.append(InstallerTreeMedia.make_scratchdir(guest))
        if self._cdrom_path():
            search_paths.append(self._cdrom_path())
        return search_paths

    def has_install_phase(self):
        """
        Return True if the requested setup is actually installing an OS
        into the guest. Things like LiveCDs, Import, or a manually specified
        bootorder do not have an install phase.
        """
        if self._no_install:
            return False
        return bool(self._cdrom or
                    self._install_bootdev or
                    self._treemedia)

    def options_specified(self):
        """
        Return True if some explicit install option was actually passed in
        Validate that some install option was actually passed in.
        """
        if self._no_install:
            return True
        return self.has_install_phase()

    def detect_distro(self, guest):
        """
        Attempt to detect the distro for the Installer's 'location'. If
        an error is encountered in the detection process (or if detection
        is not relevant for the Installer type), None is returned.

        :returns: distro variant string, or None
        """
        ret = None
        if self._treemedia:
            ret = self._treemedia.detect_distro(guest)
        elif self.cdrom:
            if guest.conn.is_remote():
                log.debug("Can't detect distro for cdrom "
                    "remote connection.")
            else:
                osguess = OSDB.guess_os_by_iso(self.cdrom)
                if osguess:
                    ret = osguess[0]
        else:
            log.debug("No media for distro detection.")

        log.debug("installer.detect_distro returned=%s", ret)
        return ret

    def set_unattended_data(self, unattended_data):
        self._unattended_data = unattended_data


    ##########################
    # guest install handling #
    ##########################

    def _prepare_get_install_xml(self, guest):
        # We do a shallow copy of the OS block here, so that we can
        # set the install time properties but not permanently overwrite
        # any config the user explicitly requested.
        data = (guest.os.bootorder, guest.os.kernel, guest.os.initrd,
                guest.os.kernel_args, guest.on_reboot, guest.currentMemory,
                guest.memory)
        return data

    def _finish_get_install_xml(self, guest, data):
        (guest.os.bootorder, guest.os.kernel, guest.os.initrd,
                guest.os.kernel_args, guest.on_reboot, guest.currentMemory,
                guest.memory) = data

    def _get_install_xml(self, guest, meter):
        data = self._prepare_get_install_xml(guest)
        try:
            self._alter_bootconfig(guest)
            self._alter_install_resources(guest, meter)
            ret = guest.get_xml()
            return ret
        finally:
            self._remove_install_cdrom_media(guest)
            self._remove_unattended_install_cdrom_device(guest)
            self._finish_get_install_xml(guest, data)

    def _build_xml(self, guest, meter):
        install_xml = None
        if self.has_install_phase():
            install_xml = self._get_install_xml(guest, meter)
        final_xml = guest.get_xml()

        log.debug("Generated install XML: %s",
            (install_xml and ("\n" + install_xml) or "None required"))
        log.debug("Generated boot XML: \n%s", final_xml)

        return install_xml, final_xml

    def _manual_transient_create(self, install_xml, final_xml, needs_boot):
        """
        For hypervisors (like vz) that don't implement createXML,
        we need to define+start, and undefine on start failure
        """
        domain = self.conn.defineXML(install_xml or final_xml)
        if not needs_boot:
            return domain

        # Handle undefining the VM if the initial startup fails
        try:
            domain.create()
        except Exception:  # pragma: no cover
            try:
                domain.undefine()
            except Exception:
                pass
            raise

        if install_xml and install_xml != final_xml:
            domain = self.conn.defineXML(final_xml)
        return domain

    def _create_guest(self, guest,
                      meter, install_xml, final_xml, doboot, transient):
        """
        Actually do the XML logging, guest defining/creating

        :param doboot: Boot guest even if it has no install phase
        """
        meter_label = _("Creating domain...")
        meter = progress.ensure_meter(meter)
        meter.start(size=None, text=meter_label)
        needs_boot = doboot or self.has_install_phase()

        if guest.type == "vz":
            if transient:
                raise RuntimeError(_("Domain type 'vz' doesn't support "
                    "transient installs."))
            domain = self._manual_transient_create(
                    install_xml, final_xml, needs_boot)

        else:
            if transient or needs_boot:
                domain = self.conn.createXML(install_xml or final_xml, 0)
            if not transient:
                domain = self.conn.defineXML(final_xml)

        try:
            log.debug("XML fetched from libvirt object:\n%s",
                          domain.XMLDesc(0))
        except Exception as e:  # pragma: no cover
            log.debug("Error fetching XML from libvirt object: %s", e)
        return domain

    def _flag_autostart(self, domain):
        """
        Set the autostart flag for domain if the user requested it
        """
        try:
            domain.setAutostart(True)
        except Exception as e:  # pragma: no cover
            if not self.conn.support.is_error_nosupport(e):
                raise
            log.warning("Could not set autostart flag: libvirt "
                            "connection does not support autostart.")


    ######################
    # Public install API #
    ######################

    def start_install(self, guest, meter=None,
                      dry=False, return_xml=False,
                      doboot=True, transient=False):
        """
        Begin the guest install. Will add install media to the guest config,
        launch it, then redefine the XML with the postinstall config.

        :param return_xml: Don't create the guest, just return generated XML
        """
        guest.validate_name(guest.conn, guest.name)
        self.set_install_defaults(guest)

        try:
            self._prepare(guest, meter)

            if not dry:
                for dev in guest.devices.disk:
                    dev.build_storage(meter)

            install_xml, final_xml = self._build_xml(guest, meter)
            if dry or return_xml:
                return (install_xml, final_xml)

            domain = self._create_guest(
                    guest, meter, install_xml, final_xml,
                    doboot, transient)

            if self.autostart:
                self._flag_autostart(domain)
            return domain
        finally:
            self._cleanup(guest)

    def get_created_disks(self, guest):
        return [d for d in guest.devices.disk if d.storage_was_created]

    def cleanup_created_disks(self, guest, meter):
        """
        Remove any disks we created as part of the install. Only ever
        called by clients.
        """
        clean_disks = self.get_created_disks(guest)
        if not clean_disks:
            return

        for disk in clean_disks:
            log.debug("Removing created disk path=%s vol_object=%s",
                disk.path, disk.get_vol_object())
            name = os.path.basename(disk.path)

            try:
                meter.start(size=None, text=_("Removing disk '%s'") % name)

                if disk.get_vol_object():
                    disk.get_vol_object().delete()
                else:  # pragma: no cover
                    # This case technically shouldn't happen here, but
                    # it's here incase future assumptions change
                    os.unlink(disk.path)

                meter.end(0)
            except Exception as e:  # pragma: no cover
                log.debug("Failed to remove disk '%s'",
                    name, exc_info=True)
                log.error("Failed to remove disk '%s': %s", name, e)



#
# Copyright 2006-2009, 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import shutil
import subprocess
import tempfile

from ..logger import log


def _run_initrd_commands(initrd, tempdir):
    log.debug("Appending to the initrd.")

    find_proc = subprocess.Popen(['find', '.', '-print0'],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 cwd=tempdir)
    cpio_proc = subprocess.Popen(['cpio', '--create', '--null', '--quiet',
                                  '--format=newc', '--owner=+0:+0'],
                                 stdin=find_proc.stdout,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 cwd=tempdir)
    f = open(initrd, 'ab')
    gzip_proc = subprocess.Popen(['gzip'], stdin=cpio_proc.stdout,
                                 stdout=f, stderr=subprocess.PIPE)
    cpio_proc.wait()
    find_proc.wait()
    gzip_proc.wait()
    f.close()

    finderr = find_proc.stderr.read()
    cpioerr = cpio_proc.stderr.read()
    gziperr = gzip_proc.stderr.read()
    if finderr:  # pragma: no cover
        log.debug("find stderr=%s", finderr)
    if cpioerr:  # pragma: no cover
        log.debug("cpio stderr=%s", cpioerr)
    if gziperr:  # pragma: no cover
        log.debug("gzip stderr=%s", gziperr)


def _run_iso_commands(iso, tempdir):
    cmd = ["genisoimage",
           "-o", iso,
           "-J",
           "-input-charset", "utf8",
           "-rational-rock",
           tempdir]
    log.debug("Running iso build command: %s", cmd)
    output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    log.debug("cmd output: %s", output)


def _perform_generic_injections(injections, scratchdir, media, cb):
    if not injections:
        return

    tempdir = tempfile.mkdtemp(dir=scratchdir)
    try:
        os.chmod(tempdir, 0o775)

        for filename in injections:
            if type(filename) is tuple:
                filename, dst = filename
            else:
                dst = os.path.basename(filename)

            log.debug("Injecting src=%s dst=%s into media=%s",
                    filename, dst, media)
            shutil.copy(filename, os.path.join(tempdir, dst))

        return cb(media, tempdir)
    finally:
        shutil.rmtree(tempdir)


def perform_initrd_injections(initrd, injections, scratchdir):
    """
    Insert files into the root directory of the initial ram disk
    """
    _perform_generic_injections(injections, scratchdir, initrd,
            _run_initrd_commands)


def perform_cdrom_injections(injections, scratchdir):
    """
    Insert files into the root directory of a generated cdrom
    """
    fileobj = tempfile.NamedTemporaryFile(
        prefix="virtinst-", suffix="-unattended.iso",
        dir=scratchdir, delete=False)
    iso = fileobj.name

    try:
        _perform_generic_injections(injections, scratchdir, iso,
            _run_iso_commands)
    except Exception:  # pragma: no cover
        os.unlink(iso)
        raise

    return iso



# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import io
import os
import unittest

from virtinst import Installer
from virtconv import VirtConverter

from tests import utils

base_dir = os.getcwd() + "/tests/virtconv-files/"
out_dir = base_dir + "libvirt_output"


class TestVirtConv(unittest.TestCase):
    def _convert_helper(self, in_path, out_path, in_type, disk_format):
        outbuf = io.StringIO()

        def print_cb(msg):
            print(msg, file=outbuf)

        conn = utils.URIs.open_kvm()
        converter = VirtConverter(conn, in_path, print_cb=print_cb)

        if converter.parser.name != in_type:
            raise AssertionError("find_parser_by_file for '%s' returned "
                                 "wrong parser type.\n"
                                 "Expected: %s\n"
                                 "Received: %s\n" %
                                 (in_path, in_type, converter.parser.name))

        converter.convert_disks(disk_format, dry=True)
        guest = converter.get_guest()
        installer = Installer(guest.conn)
        ignore, out_xml = installer.start_install(guest, return_xml=True)
        out_expect = out_xml
        if outbuf.getvalue():
            out_expect += ("\n\n" + outbuf.getvalue().replace(base_dir, ""))

        if not conn.support.conn_vmport():
            self.skipTest("Not comparing XML because vmport isn't supported")

        utils.diff_compare(out_expect, out_path)
        utils.test_create(conn, out_xml)

    def _compare(self, in_path, disk_format=None):
        in_type = "ovf"
        if "vmx" in in_path:
            in_type = "vmx"

        in_path = os.path.join(base_dir, in_path)
        base = in_type + "2libvirt"
        in_base = os.path.basename(in_path).rsplit(".", 1)[0]
        out_path = "%s/%s_%s.%s" % (out_dir, base, in_base, "libvirt")
        if disk_format:
            out_path += ".disk_%s" % disk_format

        self._convert_helper(in_path, out_path, in_type, disk_format)


    def testOVF2Libvirt(self):
        self._compare("ovf_input/test1.ovf")
        self._compare("ovf_input/test2.ovf")
        self._compare("ovf_input/test_gzip.ovf")
        self._compare("ovf_input/ovf_directory")

    def testVMX2Libvirt(self):
        self._compare("vmx_input/test1.vmx")
        self._compare("vmx_input/test-nodisks.vmx")
        self._compare("vmx_input/test-vmx-zip.zip")
        self._compare("vmx_input/vmx-dir")

    def testDiskConvert(self):
        self._compare("ovf_input/test1.ovf", disk_format="qcow2")
        self._compare("vmx_input/test1.vmx", disk_format="raw")
        self._compare("ovf_input/test_gzip.ovf", disk_format="raw")



# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import unittest

from virtinst import Interface
from tests import utils

datadir = "tests/interface-xml"


class TestInterfaces(unittest.TestCase):
    def testInterfaceBridgeIP(self):
        conn = utils.URIs.open_testdriver_cached()

        def _check_iface(name, typ, child_names):
            libvirtobj = conn.interfaceLookupByName(name)
            xmlobj = Interface(conn, parsexml=libvirtobj.XMLDesc(0))

            self.assertEqual(xmlobj.name, name)
            self.assertEqual(xmlobj.type, typ)
            self.assertEqual(
                {i.name for i in xmlobj.interfaces},
                set(child_names))

        _check_iface("eth0", "ethernet", [])
        _check_iface("bond0", "bond", ["eth-bond0-1", "eth-bond0-2"])
        _check_iface("brplain", "bridge", ["eth-brplain0", "eth-brplain1"])
        _check_iface("brempty", "bridge", [])
        _check_iface("vlaneth1.3", "vlan", ["vlaneth1"])



# Copyright (C) 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import tempfile
import unittest

import virtinst
from virtinst import DeviceDisk

from tests import utils


def _make_guest(conn=None, os_variant=None):
    if not conn:
        conn = utils.URIs.open_testdriver_cached()

    g = virtinst.Guest(conn)
    g.name = "TestGuest"
    g.currentMemory = int(200 * 1024)
    g.memory = int(400 * 1024)

    if os_variant:
        g.set_os_name(os_variant)

    # File disk
    d = virtinst.DeviceDisk(conn)
    d.path = "/dev/default-pool/new-test-suite.img"
    if d.wants_storage_creation():
        parent_pool = d.get_parent_pool()
        vol_install = virtinst.DeviceDisk.build_vol_install(conn,
            os.path.basename(d.path), parent_pool, .0000001, True)
        d.set_vol_install(vol_install)
    d.validate()
    g.add_device(d)

    # Block disk
    d = virtinst.DeviceDisk(conn)
    d.path = "/dev/disk-pool/diskvol1"
    d.validate()
    g.add_device(d)

    # Network device
    dev = virtinst.DeviceInterface(conn)
    g.add_device(dev)

    return g


def _make_installer(location=None, conn=None):
    conn = conn or utils.URIs.open_testdriver_cached()
    cdrom = not location and "/dev/null" or None
    inst = virtinst.Installer(conn, location=location, cdrom=cdrom)
    return inst


class TestXMLMisc(unittest.TestCase):
    """
    Misc tests for various XML special behavior. These should only aim for
    testing any particularly tricky bits, general XML generation should
    be through virt-install examples in clitest
    """
    @property
    def conn(self):
        return utils.URIs.open_testdefault_cached()

    def _compare(self, guest, filebase, do_install):
        filename = os.path.join("tests/xmlconfig-xml", filebase + ".xml")

        installer = _make_installer(conn=guest.conn)
        inst_xml, boot_xml = installer.start_install(
                guest, return_xml=True, dry=True)
        if do_install:
            actualXML = inst_xml
        else:
            actualXML = boot_xml

        utils.diff_compare(actualXML, filename)
        utils.test_create(guest.conn, actualXML)

    def testDiskNumbers(self):
        # Various testing our target generation
        self.assertEqual("a", DeviceDisk.num_to_target(1))
        self.assertEqual("b", DeviceDisk.num_to_target(2))
        self.assertEqual("z", DeviceDisk.num_to_target(26))
        self.assertEqual("aa", DeviceDisk.num_to_target(27))
        self.assertEqual("ab", DeviceDisk.num_to_target(28))
        self.assertEqual("az", DeviceDisk.num_to_target(52))
        self.assertEqual("ba", DeviceDisk.num_to_target(53))
        self.assertEqual("zz", DeviceDisk.num_to_target(27 * 26))
        self.assertEqual("aaa", DeviceDisk.num_to_target(27 * 26 + 1))

        self.assertEqual(DeviceDisk.target_to_num("hda"), 0)
        self.assertEqual(DeviceDisk.target_to_num("hdb"), 1)
        self.assertEqual(DeviceDisk.target_to_num("sdz"), 25)
        self.assertEqual(DeviceDisk.target_to_num("sdaa"), 26)
        self.assertEqual(DeviceDisk.target_to_num("vdab"), 27)
        self.assertEqual(DeviceDisk.target_to_num("vdaz"), 51)
        self.assertEqual(DeviceDisk.target_to_num("xvdba"), 52)
        self.assertEqual(DeviceDisk.target_to_num("xvdzz"),
            26 * (25 + 1) + 25)
        self.assertEqual(DeviceDisk.target_to_num("xvdaaa"),
            26 * 26 * 1 + 26 * 1 + 0)

        disk = virtinst.DeviceDisk(self.conn)
        disk.bus = "ide"

        self.assertEqual("hda", disk.generate_target([]))
        self.assertEqual("hdb", disk.generate_target(["hda"]))
        self.assertEqual("hdc", disk.generate_target(["hdb", "sda"]))
        self.assertEqual("hdb", disk.generate_target(["hda", "hdd"]))

        disk.bus = "scsi"
        self.assertEqual("sdb",
            disk.generate_target(["sda", "sdg", "sdi"], 0))
        self.assertEqual("sdh", disk.generate_target(["sda", "sdg"], 1))

    def testQuickTreeinfo(self):
        # Simple sanity test to make sure detect_distro works. test-urls
        # does much more exhaustive testing but it's only run occasionally
        i = _make_installer(
            location="tests/cli-test-xml/fakefedoratree")
        g = _make_guest()
        v = i.detect_distro(g)
        self.assertEqual(v, "fedora17")

        i = _make_installer(
            location="tests/cli-test-xml/fakerhel6tree")
        g = _make_guest()
        v = i.detect_distro(g)
        self.assertEqual(v, "rhel6.0")

    def testCDROMInsert(self):
        # After set_install_defaults, cdrom media should be inserted
        i = _make_installer()
        g = _make_guest()
        i.set_install_defaults(g)
        for disk in g.devices.disk:
            if disk.device == "cdrom" and disk.path == "/dev/null":
                return
        raise AssertionError("Didn't find inserted cdrom media")

    def testCPUTopology(self):
        # Test CPU topology determining
        cpu = virtinst.DomainCpu(self.conn)
        cpu.sockets = "2"
        cpu.set_topology_defaults(6)
        self.assertEqual([cpu.sockets, cpu.cores, cpu.threads], [2, 3, 1])

        cpu = virtinst.DomainCpu(self.conn)
        cpu.cores = "4"
        cpu.set_topology_defaults(9)
        self.assertEqual([cpu.sockets, cpu.cores, cpu.threads], [2, 4, 1])

        cpu = virtinst.DomainCpu(self.conn)
        cpu.threads = "3"
        cpu.set_topology_defaults(14)
        self.assertEqual([cpu.sockets, cpu.cores, cpu.threads], [4, 1, 3])

        cpu = virtinst.DomainCpu(self.conn)
        cpu.sockets = 5
        cpu.cores = 2
        self.assertEqual(cpu.vcpus_from_topology(), 10)

        cpu = virtinst.DomainCpu(self.conn)
        self.assertEqual(cpu.vcpus_from_topology(), 1)

    def test_set_defaults_double(self):
        """
        Check that a common config has idempotent set_defaults
        """
        g = _make_guest(conn=utils.URIs.open_kvm(), os_variant="fedora-unknown")

        g.set_defaults(None)
        xml1 = g.get_xml()
        g.set_defaults(None)
        xml2 = g.get_xml()
        self.assertEqual(xml1, xml2)

    def test_guest_osinfo_metadata(self):
        g = _make_guest()
        self.assertEqual(g.osinfo.name, "generic")
        g.set_os_name("fedora17")
        self.assertEqual(g.osinfo.name, "fedora17")

        g = _make_guest()
        g._metadata.libosinfo.os_id = "http://fedoraproject.org/fedora/20"  # pylint: disable=protected-access
        self.assertEqual(g.osinfo.name, "fedora20")

        g = _make_guest()
        g._metadata.libosinfo.os_id = "http://example.com/idontexit"  # pylint: disable=protected-access
        self.assertEqual(g.osinfo.name, "generic")

    @utils.run_without_testsuite_hacks
    def test_dir_searchable(self):
        # Normally the dir searchable test is skipped in the unittest,
        # but let's contrive an example that should trigger all the code
        # to ensure it isn't horribly broken
        from virtinst import diskbackend
        uid = -1
        username = "fakeuser-zzzz"
        with tempfile.TemporaryDirectory() as tmpdir:
            fixlist = diskbackend.is_path_searchable(tmpdir, uid, username)
            self.assertTrue(bool(fixlist))
            errdict = diskbackend.set_dirs_searchable(fixlist, username)
            self.assertTrue(not bool(errdict))

        import getpass
        fixlist = diskbackend.is_path_searchable(
                os.getcwd(), os.getuid(), getpass.getuser())
        self.assertTrue(not bool(fixlist))

    def test_nonpredicatble_generate(self):
        kvm_uri = utils.URIs.kvm.replace(",predictable", "")
        kvmconn = virtinst.cli.getConnection(kvm_uri)
        testconn = virtinst.cli.getConnection("test:///default")

        testuuid = virtinst.Guest.generate_uuid(self.conn)
        randomuuid = virtinst.Guest.generate_uuid(testconn)
        self.assertTrue(randomuuid != testuuid)
        self.assertTrue(len(randomuuid) == len(testuuid))

        testmac = virtinst.DeviceInterface.generate_mac(self.conn)
        randommac = virtinst.DeviceInterface.generate_mac(testconn)
        qemumac = virtinst.DeviceInterface.generate_mac(kvmconn)
        self.assertTrue(randommac != testmac)
        self.assertTrue(qemumac != testmac)
        self.assertTrue(len(randommac) == len(testmac))

    def test_support_misc(self):
        try:
            self.conn.lookupByName("foobar-idontexist")
        except Exception as e:
            if not self.conn.support.is_libvirt_error_no_domain(e):
                raise



#
# Copyright 2010, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty


class DomainFeatures(XMLBuilder):
    """
    Class for generating <features> XML
    """
    XML_NAME = "features"
    _XML_PROP_ORDER = ["acpi", "apic", "pae", "gic_version"]

    acpi = XMLProperty("./acpi", is_bool=True)
    apic = XMLProperty("./apic", is_bool=True)
    pae = XMLProperty("./pae", is_bool=True)
    gic_version = XMLProperty("./gic/@version")

    hap = XMLProperty("./hap", is_bool=True)
    viridian = XMLProperty("./viridian", is_bool=True)
    privnet = XMLProperty("./privnet", is_bool=True)

    pmu = XMLProperty("./pmu/@state", is_onoff=True)
    eoi = XMLProperty("./apic/@eoi", is_onoff=True)

    hyperv_reset = XMLProperty("./hyperv/reset/@state", is_onoff=True)
    hyperv_vapic = XMLProperty("./hyperv/vapic/@state", is_onoff=True)
    hyperv_relaxed = XMLProperty("./hyperv/relaxed/@state", is_onoff=True)
    hyperv_spinlocks = XMLProperty("./hyperv/spinlocks/@state", is_onoff=True)
    hyperv_spinlocks_retries = XMLProperty("./hyperv/spinlocks/@retries",
                                           is_int=True)
    hyperv_synic = XMLProperty("./hyperv/synic/@state", is_onoff=True)

    vmport = XMLProperty("./vmport/@state", is_onoff=True)
    kvm_hidden = XMLProperty("./kvm/hidden/@state", is_onoff=True)
    pvspinlock = XMLProperty("./pvspinlock/@state", is_onoff=True)

    smm = XMLProperty("./smm/@state", is_onoff=True)
    vmcoreinfo = XMLProperty("./vmcoreinfo/@state", is_onoff=True)


    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        if guest.os.is_container():
            self.acpi = None
            self.apic = None
            self.pae = None
            if guest.is_full_os_container() and guest.type != "vz":
                self.privnet = True
            return

        if not guest.os.is_hvm():
            return

        capsinfo = guest.lookup_capsinfo()
        if self._prop_is_unset("acpi"):
            self.acpi = capsinfo.guest.supports_acpi()
        if self._prop_is_unset("apic"):
            self.apic = capsinfo.guest.supports_apic()
        if self._prop_is_unset("pae"):
            if (guest.os.is_hvm() and
                guest.type == "xen" and
                guest.os.arch == "x86_64"):
                self.pae = True
            else:
                self.pae = capsinfo.guest.supports_pae()

        if (guest.hyperv_supported() and
            self.conn.support.conn_hyperv_vapic()):
            if self.hyperv_relaxed is None:
                self.hyperv_relaxed = True
            if self.hyperv_vapic is None:
                self.hyperv_vapic = True
            if self.hyperv_spinlocks is None:
                self.hyperv_spinlocks = True
            if self.hyperv_spinlocks_retries is None:
                self.hyperv_spinlocks_retries = 8191



#
# Copyright 2014 Fujitsu Limited.
# Chen Hanxiao <chenhanxiao at cn.fujitsu.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


class _BlkiotuneDevice(XMLBuilder):
    XML_NAME = "device"
    _XML_PROP_ORDER = ["path", "weight"]

    path = XMLProperty("./path")
    weight = XMLProperty("./weight")


class DomainBlkiotune(XMLBuilder):
    """
    Class for generating <blkiotune> XML
    """

    XML_NAME = "blkiotune"
    _XML_PROP_ORDER = ["weight"]

    weight = XMLProperty("./weight", is_int=True)
    devices = XMLChildProperty(_BlkiotuneDevice)



# Copyright (C) 2011, 2013 Red Hat, Inc.
# Copyright (C) 2011 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from virtinst import NodeDevice

from .libvirtobject import vmmLibvirtObject


def _usb_pretty_name(xmlobj):
    # Hypervisor may return a rather sparse structure, missing
    # some ol all stringular descriptions of the device altogether.
    # Do our best to help user identify the device.

    # Certain devices pad their vendor with trailing spaces,
    # such as "LENOVO       ". It does not look well.
    product = str(xmlobj.product_name).strip()
    vendor = str(xmlobj.vendor_name).strip()

    if product == "":
        product = str(xmlobj.product_id)
        if vendor == "":
            # No stringular descriptions altogether
            vendor = str(xmlobj.vendor_id)
            devstr = "%s:%s" % (vendor, product)
        else:
            # Only the vendor is known
            devstr = "%s %s" % (vendor, product)
    else:
        if vendor == "":
            # Sometimes vendor is left out empty, but product is
            # already descriptive enough or contains the vendor string:
            # "Lenovo USB Laser Mouse"
            devstr = product
        else:
            # We know everything. Perfect.
            devstr = "%s %s" % (vendor, product)

    busstr = "%.3d:%.3d" % (int(xmlobj.bus), int(xmlobj.device))
    desc = "%s %s" % (busstr, devstr)
    return desc


def _pretty_name(xmlobj):
    if xmlobj.device_type == "net":
        if xmlobj.interface:
            return _("Interface %s") % xmlobj.interface
        return xmlobj.name

    if xmlobj.device_type == "pci":
        devstr = "%.4X:%.2X:%.2X:%X" % (int(xmlobj.domain),
                                        int(xmlobj.bus),
                                        int(xmlobj.slot),
                                        int(xmlobj.function))
        return "%s %s %s" % (devstr,
                xmlobj.vendor_name, xmlobj.product_name)
    if xmlobj.device_type == "usb_device":
        return _usb_pretty_name(xmlobj)

    if xmlobj.device_type == "drm":
        parent = NodeDevice.lookupNodedevFromString(
                xmlobj.conn, xmlobj.parent)
        return "%s (%s)" % (_pretty_name(parent), xmlobj.drm_type)

    return xmlobj.name


class vmmNodeDevice(vmmLibvirtObject):
    def __init__(self, conn, backend, key):
        vmmLibvirtObject.__init__(self, conn, backend, key, NodeDevice)

    def _conn_tick_poll_param(self):
        return "pollnodedev"
    def class_name(self):
        return "nodedev"

    def _XMLDesc(self, flags):
        return self._backend.XMLDesc(flags)
    def _get_backend_status(self):
        return self._STATUS_ACTIVE
    def _backend_get_name(self):
        return self.get_connkey()
    def is_active(self):
        return True
    def _using_events(self):
        return self.conn.using_node_device_events

    def tick(self, stats_update=True):
        # Deliberately empty
        ignore = stats_update
    def _init_libvirt_state(self):
        self.ensure_latest_xml()

    def pretty_name(self):
        return _pretty_name(self.xmlobj)



#
# Copyright (C) 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

"""
Configuration variables that can be set at build time
"""

import os
import sys

if (sys.version_info.major != 3 or
    sys.version_info.minor < 4):  # pragma: no cover
    print("python 3.4 or later is required, your's is %s" %
            sys.version_info)
    sys.exit(1)

import configparser

_cfg = configparser.ConfigParser()
_filepath = os.path.abspath(__file__)
_srcdir = os.path.abspath(os.path.join(os.path.dirname(_filepath), ".."))
_cfgpath = os.path.join(os.path.dirname(_filepath), "build.cfg")
if os.path.exists(_cfgpath):
    _cfg.read(_cfgpath)  # pragma: no cover

_istest = "VIRTINST_TEST_SUITE" in os.environ
_running_from_srcdir = os.path.exists(
    os.path.join(_srcdir, "tests", "clitest.py"))


def _split_list(commastr):
    return [d for d in commastr.split(",") if d]


def _get_param(name, default):  # pragma: no cover
    if _istest:
        return default
    try:
        return _cfg.get("config", name)
    except (configparser.NoOptionError, configparser.NoSectionError):
        return default


__version__ = "2.2.1"


class _BuildConfig(object):
    def __init__(self):
        self.cfgpath = _cfgpath
        self.version = __version__

        self.default_graphics = _get_param("default_graphics", "spice")
        self.default_hvs = _split_list(_get_param("default_hvs", ""))

        self.prefix = None
        self.gettext_dir = None
        self.ui_dir = None
        self.icon_dir = None
        self.gsettings_dir = None
        self.running_from_srcdir = _running_from_srcdir
        self._set_paths_by_prefix(_get_param("prefix", "/usr"))


    def _set_paths_by_prefix(self, prefix):
        self.prefix = prefix
        self.gettext_dir = os.path.join(prefix, "share", "locale")

        if self.running_from_srcdir:
            self.ui_dir = os.path.join(_srcdir, "ui")
            self.icon_dir = os.path.join(_srcdir, "data")
            self.gsettings_dir = self.icon_dir
        else:  # pragma: no cover
            self.ui_dir = os.path.join(prefix, "share", "virt-manager", "ui")
            self.icon_dir = os.path.join(prefix, "share", "virt-manager",
                "icons")
            self.gsettings_dir = os.path.join(prefix, "share",
                "glib-2.0", "schemas")


BuildConfig = _BuildConfig()
